from libc.stdint cimport *
from libc.string cimport *

include "pyrobuf_defs.pxi"


IF PYROBUF_UNAME_SYSNAME == "Windows":
    cdef extern from "strndup.c":
        char * strndup (const char *s, size_t n)


cdef safe_fill(char *dest, const unsigned char *source, size_t max_len):
    """
    Fill preallocated buffer "dest" from string "source" up to length max_len,
    check for string null values, and append a null terminator.
    """
    cdef const char *cast = <const char *>source
    if max_len == 0 or cast == NULL or cast[0] == b'\0':
        dest[0] = b'\0'
        return

    if max_len >= 4 and str(cast).upper() == "NULL":
        dest[0] = b'\0'
        return

    strncpy(dest, cast, max_len - 1)
    dest[max_len - 1] = b'\0'


cdef safe_dup(const unsigned char *source, size_t max_len):
    """
    Return a new string duplicated from string "source" up to length max_len,
    and with a null terminator appended.
    """
    cdef const char *cast = <const char *>source
    if max_len == 0 or cast == NULL or cast[0] == b'\0':
        return b''

    s = b'' + strndup(cast, max_len)
    if max_len >= 4 and str(s).upper() == "NULL":
        return b''

    return s


cdef int32_t get_varint32(const unsigned char *varint, int *offset):
    """
    Deserialize a protobuf varint starting from give offset in memory; update
    offset based on number of bytes consumed.
    """
    cdef int32_t value = 0
    cdef int32_t base = 1
    cdef int index = 0
    cdef int val_byte

    while True:
        val_byte = varint[offset[0] + index]
        value += (val_byte & 0x7F) * base
        if (val_byte & 0x80):
            base *= 128
            index += 1
        else:
            offset[0] += (index + 1)
            return value


cdef int64_t get_varint64(const unsigned char *varint, int *offset):
    """
    Deserialize a protobuf varint starting from give offset in memory; update
    offset based on number of bytes consumed.
    """
    cdef int64_t value = 0
    cdef int64_t base = 1
    cdef int index = 0
    cdef int val_byte

    while True:
        val_byte = varint[offset[0] + index]
        value += (val_byte & 0x7F) * base
        if (val_byte & 0x80):
            base *= 128
            index += 1
        else:
            offset[0] += (index + 1)
            return value

def get_varint(data, offset=0):
    """
    Return an integer value obtained by decoding varint data from the given
    byte string beginning at the specified offset.
    """
    cdef int _offset = offset
    return get_varint64(data, &_offset)


cdef int32_t get_signed_varint32(const unsigned char *varint, int *offset):
    """
    Deserialize a signed protobuf varint starting from give offset in memory;
    update offset based on number of bytes consumed.
    """
    cdef uint32_t value = 0
    cdef int32_t base = 1
    cdef int index = 0
    cdef int val_byte

    while True:
        val_byte = varint[offset[0] + index]
        value += (val_byte & 0x7F) * base
        if (val_byte & 0x80):
            base *= 128
            index += 1
        else:
            offset[0] += (index + 1)
            return <int32_t>((value >> 1) ^ (-(value & 1))) # zigzag decoding

cdef int64_t get_signed_varint64(const unsigned char *varint, int *offset):
    """
    Deserialize a signed protobuf varint starting from give offset in memory;
    update offset based on number of bytes consumed.
    """
    cdef uint64_t value = 0
    cdef int64_t base = 1
    cdef int index = 0
    cdef int val_byte

    while True:
        val_byte = varint[offset[0] + index]
        value += (val_byte & 0x7F) * base
        if (val_byte & 0x80):
            base *= 128
            index += 1
        else:
            offset[0] += (index + 1)
            return <int64_t>((value >> 1) ^ (-(value & 1))) # zigzag decoding

def get_signed_varint(data, offset=0):
    """
    Return an integer value obtained by decoding signed-varint data from the
    given byte string beginning at the specified offset.
    """
    cdef int _offset = offset
    return get_signed_varint64(data, &_offset)


cdef int set_varint32(int32_t varint, bytearray buf):
    """
    Serialize an integer into a protobuf varint; return the number of bytes
    serialized.
    """

	# Negative numbers are always 10 bytes, so we need a uint64_t to
    # facilitate encoding
    cdef uint64_t enc = varint
    cdef uint8_t bits = enc & 0x7f
    enc >>= 7
    cdef int idx = 1
    while enc:
        buf.append(<unsigned char>(0x80|bits))
        bits = enc & 0x7f
        enc >>= 7
        idx += 1
    buf.append(<unsigned char>bits)
    return idx + 1

cdef int set_varint64(int64_t varint, bytearray buf):
    """
    Serialize an integer into a protobuf varint; return the number of bytes
    serialized.
    """

    # Negative numbers are always 10 bytes, so we need a uint64_t to
    # facilitate encoding
    cdef uint64_t enc = varint
    cdef uint8_t bits = enc & 0x7f
    enc >>= 7
    cdef int idx = 1
    while enc:
        buf.append(<unsigned char>(0x80|bits))
        bits = enc & 0x7f
        enc >>= 7
        idx += 1
    buf.append(<unsigned char>bits)
    return idx + 1

def to_varint(varint):
    """
    Return a byte string containing the varint encoded form of the specified
    varint.
    """
    buf = bytearray()
    set_varint64(varint, buf)
    return buf


cdef int set_signed_varint32(int32_t varint, bytearray buf):
    """
    Serialize an integer into a signed protobuf varint; return the number of
    bytes serialized.
    """
    cdef uint32_t enc
    cdef uint8_t bits
    cdef int idx = 1

    enc = (varint << 1) ^ (varint >> 31) # zigzag encoding
    bits = enc & 0x7f
    enc >>= 7
    while enc:
        buf.append(<unsigned char>(bits | 0x80))
        bits = enc & 0x7f
        enc >>= 7
        idx += 1

    buf.append(<unsigned char>bits)
    return idx + 1


cdef int set_signed_varint64(int64_t varint, bytearray buf):
    """
    Serialize an integer into a signed protobuf varint; return the number of
    bytes serialized.
    """
    cdef uint64_t enc
    cdef uint8_t bits
    cdef int idx = 1

    enc = (varint << 1) ^ (varint >> 63) # zigzag encoding
    bits = enc & 0x7f
    enc >>= 7
    while enc:
        buf.append(<unsigned char>(bits | 0x80))
        bits = enc & 0x7f
        enc >>= 7
        idx += 1

    buf.append(<unsigned char>bits)
    return idx + 1


def to_signed_varint(varint):
    """
    Return a byte string containing the signed-varint encoded form of the
    specified varint.
    """
    buf = bytearray()
    set_signed_varint64(varint, buf)
    return buf


def from_varint(data, offset=0):
    """
    Return a (integer value, new_offset) pair obtained by decoding varint data
    from the given byte string beginning at the specified offset.
    """
    cdef int _offset = offset
    result = get_varint64(data, &_offset)
    return result, _offset


def from_signed_varint(data, offset=0):
    """
    Return a (integer value, new_offset) pair obtained by decoding signed varint
    data from the given byte string beginning at the specified offset.
    """
    cdef int _offset = offset
    result = get_signed_varint64(data, &_offset)
    return result, _offset


cdef bint skip_generic(const unsigned char *memory, int *offset, int size, int wire_type):
    """
    Parse field of given wire type to update offset.
    """
    cdef int64_t skip

    if wire_type == 0:
        get_varint64(memory, offset)
    elif wire_type == 1:
        offset[0] += 8
    elif wire_type == 2:
        skip = get_varint64(memory, offset)
        offset[0] += skip
    elif wire_type == 5:
        offset[0] += 4
    else:
        return False

    return offset[0] <= size



import argparse
import glob
import os
import sys
from setuptools import setup

from Cython.Build import cythonize
from jinja2 import Environment, PackageLoader

from pyrobuf.parse_proto import Parser, Proto3Parser

if sys.version_info.major == 3:
    _FileExistsError = FileExistsError
else:
    _FileExistsError = OSError

_VM = sys.version_info.major


class Compiler(object):

    _env = Environment(loader=PackageLoader('pyrobuf.protobuf', 'templates'))
    t_pxd = _env.get_template('proto_pxd.tmpl')
    t_pyx = _env.get_template('proto_pyx.tmpl')

    def __init__(self, source, out="out", build="build", install=False,
                 proto3=False, force=False, package=None):
        self.source = source
        self.out = out
        self.build = build
        self.install = install
        self.force = force
        self.package = package
        here = os.path.dirname(os.path.abspath(__file__))
        self.include_path = [os.path.join(here, 'src'), self.out]
        self._generated = set()
        self._messages = []
        self._pyx_files = []

        if proto3:
            self.parser = Proto3Parser
        else:
            self.parser = Parser

    @classmethod
    def parse_cli_args(cls):
        parser = argparse.ArgumentParser(
            "pyrobuf", description="a Cython based protobuf compiler")
        parser.add_argument('source', type=str,
                            help="<filename>.proto or directory containing "
                                 "proto files")
        parser.add_argument('--out-dir', default='out',
                            help="cythonize output directory [default: out]")
        parser.add_argument('--build-dir', default='build',
                            help="C compiler build directory [default: build]")
        parser.add_argument('--install', action='store_true',
                            help="install the extension [default: False]")
        parser.add_argument('--proto3', action='store_true',
                            help="compile proto3 syntax [default: False]")
        parser.add_argument('--force', action='store_true',
                            help="force install")
        parser.add_argument('--package', type=str, default=None,
                            help="name of package to compile to")
        args = parser.parse_args()

        return cls(args.source, out=args.out_dir, build=args.build_dir,
                   install=args.install, proto3=args.proto3, force=args.force,
                   package=args.package)

    def compile(self):
        script_args = ['build', '--build-base={0}'.format(self.build)]

        if self.install:
            script_args.append('install')

        if self.force:
            script_args.append('--force')

        self._compile_spec()

        if self.package is not None:
            self._package()

        setup(name='pyrobuf-generated',
              ext_modules=cythonize(self._pyx_files,
                                    include_path=self.include_path),
              script_args=script_args)

    def extend(self, dist):
        self._compile_spec()

        if self.package is not None:
            self._package()

        if dist.ext_modules is None:
            dist.ext_modules = []

        dist.ext_modules.extend(cythonize(self._pyx_files,
                                          include_path=self.include_path))

    def _compile_spec(self):
        try:
            os.makedirs(self.out)
        except _FileExistsError:
            pass

        if os.path.isdir(self.source):
            for spec in glob.glob(os.path.join(self.source, '*.proto')):
                self._generate(spec)
        else:
            self._generate(self.source)

    def _generate(self, filename):
        name, _ = os.path.splitext(os.path.basename(filename))
        directory = os.path.dirname(filename)

        if name in self._generated:
            return

        print("generating {0}".format(filename))
        self._generated.add(name)

        msg_def = self.parser.parse_from_filename(filename)
        self._messages.append(msg_def)

        for f in msg_def['imports']:
            print("parsing dependency '{}'".format(f))
            try:
                self._generate(os.path.join(directory, '{}.proto'.format(f)))
            except FileNotFoundError:
                raise FileNotFoundError("can't find message spec for '{}'"
                                        .format(f))

        if self.package is None:
            self._write(name, msg_def)

    def _write(self, name, msg_def):
        name_pxd = "{}_proto.pxd".format(name)
        name_pyx = "{}_proto.pyx".format(name)
        self._pyx_files.append(os.path.join(self.out, name_pyx))

        with open(os.path.join(self.out, name_pxd), 'w') as fp:
            fp.write(self.t_pxd.render(msg_def, version_major=_VM))

        with open(os.path.join(self.out, name_pyx), 'w') as fp:
            fp.write(self.t_pyx.render(msg_def, version_major=_VM))

    def _package(self):
        meta = {'imports': [], 'messages': [], 'enums': []}

        for msg_def in self._messages:
            meta['messages'].extend(msg_def['messages'])
            meta['enums'].extend(msg_def['enums'])

        name_pxd = "{}.pxd".format(self.package)
        name_pyx = "{}.pyx".format(self.package)
        self._pyx_files.append(os.path.join(self.out, name_pyx))

        with open(os.path.join(self.out, name_pxd), 'w') as fp:
            fp.write(self.t_pxd.render(meta, version_major=_VM))

        with open(os.path.join(self.out, name_pyx), 'w') as fp:
            fp.write(self.t_pyx.render(meta, version_major=_VM))



from libc.stdint cimport *
from cpython.ref cimport PyObject
from cpython cimport Py_buffer


cdef class NullListener:
    cpdef void _Modified(self)

cdef public NullListener null_listener


cdef class TypedList(list):
    
    cdef type _list_type
    cdef PyObject *_listener


cdef class BytesList(list):

    cdef PyObject *_listener


cdef class StringList(list):

    cdef PyObject *_listener






cdef class DoubleList:

    cdef double *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, double x)

    cpdef _equal(self, other)
    cpdef append(self, double x)
    cpdef extend(self, x)
    cpdef insert(self, int i, double x)
    cpdef pop(self)
    cpdef remove(self, double x)




cdef class FloatList:

    cdef float *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, float x)

    cpdef _equal(self, other)
    cpdef append(self, float x)
    cpdef extend(self, x)
    cpdef insert(self, int i, float x)
    cpdef pop(self)
    cpdef remove(self, float x)




cdef class IntList:

    cdef int *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, int x)

    cpdef _equal(self, other)
    cpdef append(self, int x)
    cpdef extend(self, x)
    cpdef insert(self, int i, int x)
    cpdef pop(self)
    cpdef remove(self, int x)




cdef class Int32List:

    cdef int32_t *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, int32_t x)

    cpdef _equal(self, other)
    cpdef append(self, int32_t x)
    cpdef extend(self, x)
    cpdef insert(self, int i, int32_t x)
    cpdef pop(self)
    cpdef remove(self, int32_t x)




cdef class Uint32List:

    cdef uint32_t *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, uint32_t x)

    cpdef _equal(self, other)
    cpdef append(self, uint32_t x)
    cpdef extend(self, x)
    cpdef insert(self, int i, uint32_t x)
    cpdef pop(self)
    cpdef remove(self, uint32_t x)




cdef class Int64List:

    cdef int64_t *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, int64_t x)

    cpdef _equal(self, other)
    cpdef append(self, int64_t x)
    cpdef extend(self, x)
    cpdef insert(self, int i, int64_t x)
    cpdef pop(self)
    cpdef remove(self, int64_t x)




cdef class Uint64List:

    cdef uint64_t *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, uint64_t x)

    cpdef _equal(self, other)
    cpdef append(self, uint64_t x)
    cpdef extend(self, x)
    cpdef insert(self, int i, uint64_t x)
    cpdef pop(self)
    cpdef remove(self, uint64_t x)




cdef class CharList:

    cdef char *_data
    cdef size_t _n_items
    cdef size_t _size
    cdef PyObject *_listener
    cdef Py_ssize_t shape[1]
    cdef Py_ssize_t strides[1]

    cdef void _append(self, char x)

    cpdef _equal(self, other)
    cpdef append(self, char x)
    cpdef extend(self, x)
    cpdef insert(self, int i, char x)
    cpdef pop(self)
    cpdef remove(self, char x)





from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from libc.stdint cimport *
from cpython.ref cimport PyObject
from cpython.object cimport Py_EQ, Py_NE


cdef class NullListener:
    property _is_present_in_parent:
        def __get__(self):
            return True

    cpdef void _Modified(self):
        return

cdef public NullListener null_listener = NullListener()


cdef class TypedList(list):

    def __init__(self, type list_type, object listener=None):
        self._list_type = list_type
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    property list_type:
        def __get__(self):
            return self._list_type

    def __setitem__(self, i, x):
        try:
            assert isinstance(x, self._list_type)
        except AssertionError:
            raise Exception("type mismatch")

        super(TypedList, self).__setitem__(i, x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()



    def add(self, **kwargs):
        elt = self._list_type(**kwargs)
        super(TypedList, self).append(elt)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return elt

    def append(self, x):
        try:
            assert isinstance(x, self._list_type)
        except AssertionError:
            raise Exception("type mismatch")

        super(TypedList, self).append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def extend(self, x):
        for i in x:
            elt = self._list_type()
            elt.MergeFrom(i)
            super(TypedList, self).append(elt)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def insert(self, i, x):
        try:
            assert isinstance(x, self._list_type)
        except AssertionError:
            raise Exception("type mismatch")

        super(TypedList, self).insert(i, x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()


cdef class BytesList(list):

    def __init__(self, object listener=None):
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __setitem__(self, i, x):
        cdef list values

        if isinstance(i, int):
            if not isinstance(x, bytes):
                raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))

            list.__setitem__(self, i, x)
        elif isinstance(i, slice):
            values = list()

            for val in x:
                if isinstance(val, bytes):
                    values.append(val)
                else:
                    raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))

            list.__setitem__(self, i, values)
        else:
            raise TypeError("list indices must be integers, not %s" % type(i).__name__)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()



    def append(self, x):
        if not isinstance(x, bytes):
            raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))

        list.append(self, x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def extend(self, x):
        for val in x:
            if not isinstance(val, bytes):
                raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
            list.append(self, val)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def insert(self, i, x):

        if not isinstance(x, bytes):
            raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))

        list.insert(self, i, x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()


cdef class StringList(list):

    def __init__(self, object listener=None):
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __setitem__(self, i, x):
        cdef list values
        cdef str value

        if isinstance(i, int):
            if isinstance(x, str):
                value = x
            elif isinstance(x, bytes):
                value = x.decode('utf-8')
            else:
                raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))

            list.__setitem__(self, i, value)
        elif isinstance(i, slice):
            values = list()

            for val in x:
                if isinstance(val, str):
                    values.append(val)
                elif isinstance(val, bytes):
                    values.append(val.decode('utf-8'))
                else:
                    raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))

            list.__setitem__(self, i, values)
        else:
            raise TypeError("list indices must be integers, not %s" % type(i).__name__)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()



    def append(self, x):
        cdef str value
        if isinstance(x, str):
            value = x
        elif isinstance(x, bytes):
            value = x.decode('utf-8')
        else:
            raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))

        list.append(self, value)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def extend(self, x):
        cdef str value

        for val in x:
            if isinstance(val, bytes):
                value = val.decode('utf-8')
            elif isinstance(val, str):
                value = val
            else:
                raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))
            list.append(self, value)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def insert(self, i, x):
        cdef str value
        if isinstance(x, bytes):
            value = x.decode('utf-8')
        elif isinstance(x, str):
            value = x
        else:
            raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))

        list.insert(self, i, value)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()






cdef class DoubleList:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <double *>PyMem_Malloc(size * sizeof(double))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, double x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, double x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(double)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'd'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, double x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, double x):
        cdef double *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, double x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, double x):
        cdef double *mem

        if self._n_items == self._size:
            mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class FloatList:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <float *>PyMem_Malloc(size * sizeof(float))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, float x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, float x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(float)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'f'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, float x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, float x):
        cdef float *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, float x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, float x):
        cdef float *mem

        if self._n_items == self._size:
            mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class IntList:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <int *>PyMem_Malloc(size * sizeof(int))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, int x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, int x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(int)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'i'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, int x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, int x):
        cdef int *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, int x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, int x):
        cdef int *mem

        if self._n_items == self._size:
            mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class Int32List:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, int32_t x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, int32_t x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(int32_t)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'i'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, int32_t x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, int32_t x):
        cdef int32_t *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, int32_t x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, int32_t x):
        cdef int32_t *mem

        if self._n_items == self._size:
            mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class Uint32List:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, uint32_t x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, uint32_t x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(uint32_t)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'I'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, uint32_t x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, uint32_t x):
        cdef uint32_t *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, uint32_t x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, uint32_t x):
        cdef uint32_t *mem

        if self._n_items == self._size:
            mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class Int64List:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, int64_t x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, int64_t x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(int64_t)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'L'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, int64_t x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, int64_t x):
        cdef int64_t *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, int64_t x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, int64_t x):
        cdef int64_t *mem

        if self._n_items == self._size:
            mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class Uint64List:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, uint64_t x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, uint64_t x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(uint64_t)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'K'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, uint64_t x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, uint64_t x):
        cdef uint64_t *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, uint64_t x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, uint64_t x):
        cdef uint64_t *mem

        if self._n_items == self._size:
            mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1



cdef class CharList:

    def __cinit__(self, size_t size=16, object listener=None):
        self._data = <char *>PyMem_Malloc(size * sizeof(char))
        if not self._data:
            raise MemoryError()

        self._n_items = 0
        self._size = size
        if listener is None:
            self._listener = <PyObject *>null_listener
        else:
            self._listener = <PyObject *>listener

    def __dealloc__(self):
        PyMem_Free(self._data)

    def __contains__(self, char x):
        cdef size_t i
        for i in range(self._n_items):
            if self._data[i] == x:
                return True

        return False

    def __delitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        cdef size_t j
        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __getitem__(self, int i):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items or i < 0:
            raise IndexError("list index out of range")

        return self._data[i]

    def __iter__(self):
        cdef size_t idx = 0
        while idx < self._n_items:
            yield self._data[idx]
            idx += 1

    def __len__(self):
        return self._n_items

    def __repr__(self):
        return repr(list(self))

    def __setitem__(self, int i, char x):
        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i >= self._n_items:
            raise IndexError("list index out of range")

        self._data[i] = x
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    def __str__(self):
        return str(list(self))

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        
        cdef Py_ssize_t itemsize = sizeof(char)
        self.shape[0] = self._n_items
        self.strides[0] = 1

        buffer.buf = <char *>(self._data)
        buffer.format = 'c'
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = self._n_items
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 1
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __richcmp__(self, other, op):
        if op == Py_EQ:
            return self._equal(other)
        elif op == Py_NE:
            return not self._equal(other)
        else:
            raise TypeError('Operation not supported for lists')

    def __neq__(self, other):
        return not self == other

    cpdef _equal(self, other):
        try:
            if len(self) != len(other):
                return False
        except TypeError:
            return False

        for x, y in zip(self, other):
            if x != y:
                return False

        return True

    cpdef append(self, char x):
        self._append(x)
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef extend(self, x):
        for i in x:
            self._append(i)

        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef insert(self, int i, char x):
        cdef char *mem
        cdef size_t j

        if i < 0:
            i += <int>self._n_items

        if i < 0 or <size_t>i > self._n_items:
            raise IndexError("list index out of range")

        if self._n_items == self._size:
            mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        for j in range(self._n_items, <size_t>i, -1):
            self._data[j] = self._data[j - 1]

        self._data[i] = x
        self._n_items += 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cpdef pop(self):
        if self._n_items == 0:
            raise IndexError("pop from empty list")

        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()
        return self._data[self._n_items]

    cpdef remove(self, char x):
        cdef size_t i
        cdef size_t j
        cdef bint found = 0

        for i in range(self._n_items):
            if self._data[i] == x:
                found = 1
                break

        if found == 0:
            raise ValueError("x not in list")

        for j in range(i, self._n_items):
            self._data[j] = self._data[j + 1]
        
        self._n_items -= 1
        listener = <object>self._listener
        if not listener._is_present_in_parent:
            listener._Modified()

    cdef void _append(self, char x):
        cdef char *mem

        if self._n_items == self._size:
            mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
            if not mem:
                raise MemoryError()

            self._data = mem
            self._size *= 2

        self._data[self._n_items] = x
        self._n_items += 1




import os
import re


class Parser(object):

    syntax = 2

    # Tokens ordered roughly by priority
    tokens = [
        ('COMMENT_OL', r'\/\/.*?\n'),
        ('COMMENT_ML', r'\/\*(?:.|[\r\n])*?\*\/'),
        ('OPTION', r'option\s+((?:.|[\n\r])*?);'),
        ('IMPORT', r'import\s+"(.+?).proto"\s*;'),
        ('MESSAGE', r'message\s+([A-Za-z_][0-9A-Za-z_]*)'),
        ('ENUM', r'enum\s+([A-Za-z_][0-9A-Za-z_]*)'),
        ('PACKAGE', r'package\s+[A-Za-z_][0-9A-Za-z_\.]*\s*;'),
        ('SYNTAX', r'syntax\s*=\s*"(.*?)"\s*;'),
        ('EXTEND', r'extend\s+([A-Za-z_][0-9A-Za-z_\.]*)'),
        ('EXTENSION', r'extensions\s+(\d+)\s+to\s+(\d+|max)\s*;'),
        ('ONEOF', r'oneof\s+([A-Za-z_][0-9A-Za-z_]*)'),
        ('MODIFIER', r'(optional|required|repeated)'),
        ('FIELD', r'([A-Za-z][0-9A-Za-z_]*)\s+([A-Za-z][0-9A-Za-z_]*)\s*='
                  r'\s*(\d+)'),
        ('DEFAULT', r'default\s*=\s*([A-Za-z][0-9A-Za-z_]*|-?[0-9]*\.?[0-9]+'
                    r'(?:[eE][-+]?[0-9]+)?|"(?:[^"\\]|\\.)*"|\'(?:[^\'\\]|\\.)*'
                    r'\')'),
        ('PACKED', r'packed\s*=\s*(true|false)'),
        ('DEPRECATED', r'deprecated\s*=\s*(true|false)'),
        ('GENERIC', r'(\([A-Za-z][0-9A-Za-z_]*\).[A-Za-z][0-9A-Za-z_]*)\s*='
                    r'\s*([^\]]+)'),
        ('LBRACKET', r'\['),
        ('RBRACKET', r'\]\s*;'),
        ('LBRACE', r'\{'),
        ('RBRACE', r'\}\s*;{0,1}'),
        ('COMMA', r','),
        ('SKIP', r'\s'),
        ('SEMICOLON', r';'),
        ('ENUM_FIELD_WITH_VALUE', r'([A-Za-z_][0-9A-Za-z_]*)\s*='
                                  r'\s*(0x[0-9A-Fa-f]+|-\d+|\d+)'),
        ('ENUM_FIELD', r'([A-Za-z_][0-9A-Za-z_]*)')
    ]

    scalars = {
        'double', 'float', 'int32', 'int64', 'uint32', 'uint64', 'sint32',
        'sint64', 'fixed32', 'fixed64', 'sfixed32', 'sfixed64', 'bool', 'enum',
        'timestamp'
    }

    list_type_map = {
        'float':    'FloatList',
        'double':   'DoubleList',
        'enum':     'Int32List',
        'int32':    'Int32List',
        'sint32':   'Int32List',
        'sfixed32': 'Int32List',
        'uint32':   'Uint32List',
        'fixed32':  'Uint32List',
        'bool':     'Uint32List',
        'int64':    'Int64List',
        'sint64':   'Int64List',
        'sfixed64': 'Int64List',
        'uint64':   'Uint64List',
        'fixed64':  'Uint64List',
        'string':   'StringList',
        'bytes':    'BytesList'
    }

    c_type_map = {
        'float':    'float',
        'double':   'double',
        'int32':    'int32_t',
        'sint32':   'int32_t',
        'sfixed32': 'int32_t',
        'uint32':   'uint32_t',
        'fixed32':  'uint32_t',
        'int64':    'int64_t',
        'sint64':   'int64_t',
        'sfixed64': 'int64_t',
        'uint64':   'uint64_t',
        'fixed64':  'uint64_t',
        'bool':     'uint32_t',
        'enum':     'int32_t',
        'timestamp':'uint32_t'
    }

    getter_map = {
        'bool':     'get_varint32',
        'enum':     'get_varint32',
        'int32':    'get_varint32',
        'sint32':   'get_signed_varint32',
        'uint32':   'get_varint32',
        'int64':    'get_varint64',
        'sint64':   'get_signed_varint64',
        'uint64':   'get_varint64'
    }

    setter_map = {
        'bool':     'set_varint32',
        'enum':     'set_varint32',
        'int32':    'set_varint32',
        'sint32':   'set_signed_varint32',
        'uint32':   'set_varint32',
        'int64':    'set_varint64',
        'sint64':   'set_signed_varint64',
        'uint64':   'set_varint64'
    }

    token_regex = '|'.join('(?P<%s>%s)' % pair for pair in tokens)
    get_token = re.compile(token_regex).match
    token_getter = {key: re.compile(val).match for key, val in tokens}

    def __init__(self, string):
        self.string = string
        self.lines = string.split('\n')

    def tokenize(self):
        pos = 0
        line = 0

        m = self.get_token(self.string, pos)
        while m is not None:
            token_type = m.lastgroup
            subm = self.token_getter[token_type](m.group(token_type))
            vals = subm.groups()

            if token_type == 'OPTION':
                yield self.Option(line, *vals)

            elif token_type == 'SYNTAX':
                yield self.Syntax(line, *vals)

            elif token_type == 'IMPORT':
                yield self.Import(line, *vals)

            elif token_type == 'MESSAGE':
                yield self.Message(line, *vals)

            elif token_type == 'MODIFIER':
                yield self.Modifier(line, *vals)

            elif token_type == 'FIELD':
                yield self.Field(line, *vals)

            elif token_type == 'LBRACKET':
                yield self.LBracket(line)

            elif token_type == 'RBRACKET':
                yield self.RBracket(line)

            elif token_type == 'DEFAULT':
                yield self.Default(line, *vals)

            elif token_type == 'PACKED':
                yield self.Packed(line, *vals)

            elif token_type == 'DEPRECATED':
                yield self.Deprecated(line, *vals)

            elif token_type == 'CUSTOM':
                yield self.Custom(line, *vals)

            elif token_type == 'COMMA':
                yield self.Comma(line)

            elif token_type == 'SEMICOLON':
                yield self.Semicolon(line)

            elif token_type == 'ENUM':
                yield self.Enum(line, *vals)

            elif token_type in {'ENUM_FIELD', 'ENUM_FIELD_WITH_VALUE'}:
                yield self.EnumField(line, *vals)

            elif token_type == 'LBRACE':
                yield self.LBrace(line)

            elif token_type == 'RBRACE':
                yield self.RBrace(line)

            elif token_type == 'EXTENSION':
                yield self.Extension(line, *vals)

            elif token_type == 'ONEOF':
                yield self.Oneof(line, *vals)

            elif token_type == 'EXTEND':
                yield self.Extend(line, *vals)

            line += subm.group().count('\n')
            pos = m.end()
            m = self.get_token(self.string, pos)

        if pos != len(self.string):
            raise Exception("Unexpected character '{}' on line {}: '{}'".format(
                self.string[pos], line + 1, self.lines[line]))

    def parse(self, cython_info=True, fname=''):
        tokens = self.tokenize()
        rep = {'imports': [], 'messages': [], 'enums': []}
        enums = {}
        imported = {'messages': {}, 'enums': {}}
        messages = {}

        for token in tokens:
            if token.token_type == 'OPTION':
                continue

            elif token.token_type == 'SYNTAX':
                if 'proto2' == token.value:
                    self.syntax = 2
                elif 'proto3' == token.value:
                    self.syntax = 3
                else:
                    raise Exception("Unexpected syntax value '{}'".format(
                        token.value))

            elif token.token_type == 'IMPORT':
                # Ignore google meta messages
                if token.value.find('google/protobuf') == 0:
                    continue

                rep['imports'].append(token.value)

                # Google's protoc only supports the use of messages and enums
                # from direct imports. So messages and enums from indirect
                # imports are not fetched here.
                imported_rep = self._parse_import(token.value + '.proto', fname)
                imported['messages'].update((m.name, m) for m in imported_rep['messages'])
                imported['enums'].update((e.name, e) for e in imported_rep['enums'])

            elif token.token_type == 'MESSAGE':
                rep['messages'].append(self._parse_message(token, tokens, messages, enums.copy(), imported['enums']))

            elif token.token_type == 'ENUM':
                ret = self._parse_enum(token, tokens)
                rep['enums'].append(ret)
                enums[token.name] = token

            elif token.token_type == 'EXTEND':
                self._parse_extend(token, tokens)

            else:
                raise Exception("unexpected %s token on line %d: '%s'" % (
                    token.type, token.line + 1, self.lines[token.line]))

        if cython_info:
            for message in rep['messages']:
                self.add_cython_info(message)

        return rep

    @classmethod
    def parse_from_filename(cls, fname):
        with open(fname, 'r') as fp:
            s = fp.read()

        try:
            return cls(s).parse(fname=fname)
        except Exception as e:
            print('Exception while parsing {}'.format(fname))
            raise e

    def _parse_import(self, fname, parent_fname):
        actual_fname = fname if os.path.isabs(fname) else os.path.join(os.path.dirname(parent_fname), fname)
        rep = self.__class__.parse_from_filename(actual_fname)
        return rep

    def _process_token_enum(self, token, enums):
        """
        Helper method for processing a token using known enums.
            * replaces token.default enum string value
              by its equivalent enum int value
            * sets token.enum_default attribute to enum default string value
            * sets few other token attributes
        Args:
            token: a ParserField.
            enums: {string: ParserEnum} dictionary.
        """
        if token.default is not None:
            default = None
            enum_default = None
            found = False
            for entry in enums[token.type].fields:
                if token.default == entry.name:
                    default = entry.value
                    enum_default = entry.full_name
                    found = True
                    break

            assert found, ('Enum type "{}" has no value named "{}" on line {}: '
                           '{}'.format(token.type, token.default,
                                       token.line + 1, self.lines[token.line]))
            token.default = default
            token.enum_default = enum_default

        token.enum_def = enums[token.type]
        token.enum_name = token.enum_def.full_name
        token.type = 'enum'

    def _parse_message(self, current, tokens, messages, enums, imported_enums):
        """
        Recursive parsing of messages.
        Args:
            s: the proto content string.
            current: the current ParserMessage object we are working on.
            tokens: a generator of Parser*.
            messages: a dictionary of all ParserMessage objects already known/parsed.
            enums: a dictionary of ParserEnum objects.
            imported_enums: a dictionary of ParserEnum objects.
        Returns:
            a list/hiearchy of ParserMessage objects.
        """
        token = next(tokens)
        previous = self.LBrace(-1)

        assert token.token_type == 'LBRACE', (
            "missing opening brace on line {}: '{}'".format(
                token.line + 1, self.lines[token.line]))

        for token in tokens:
            if token.token_type == 'MESSAGE':
                token.full_name = current.full_name + token.name
                current.messages[token.name] = self._parse_message(token, tokens, messages, enums.copy(), imported_enums)
                # updates the dictionary of known/parsed messages.
                messages[token.name] = current.messages[token.name]

            elif token.token_type == 'ENUM':
                token.full_name = current.full_name + token.name
                current.enums[token.name] = self._parse_enum(token, tokens)
                # updates the dictionary of known/parsed enums
                enums[token.name] = current.enums[token.name]

            elif token.token_type == 'MODIFIER':
                if self.syntax == 3:
                    assert token.value == 'repeated', (
                        "Illegal modifier {} on line {}: {}".format(
                            token.value, token.line + 1,
                            self.lines[token.line]))

            elif token.token_type == 'FIELD':
                if self.syntax == 2:
                    assert previous.token_type == 'MODIFIER', (
                        "Need modifier for field on line {}: {}".format(
                            token.line + 1, self.lines[token.line]))

                if previous.token_type == 'MODIFIER':
                    token.modifier = previous.value

                self._parse_field(token, tokens)

                if messages.get(token.type) is not None:
                    # retrieves the type "full_name"
                    token.message_name = messages.get(token.type).full_name
                    token.type = 'message'

                elif current.enums.get(token.type) is not None:
                    self._process_token_enum(token, current.enums)

                elif enums.get(token.type) is not None:
                    self._process_token_enum(token, enums)

                elif imported_enums.get(token.type) is not None:
                    self._process_token_enum(token, imported_enums)

                elif (token.type not in self.scalars) and (
                            token.type not in {'string', 'bytes'}):
                    token.message_name = token.type
                    token.type = 'message'

                current.fields.append(token)

            elif token.token_type == 'EXTENSION':
                # Just ignore extensions for now, but don't error
                previous = token
                continue

            elif token.token_type == 'OPTION':
                # Just ignore options for now, but don't error
                previous = token
                continue

            else:
                assert token.token_type == 'RBRACE', (
                    "unexpected {} token on line {}: '{}'".format(
                        token.token_type, token.line + 1,
                        self.lines[token.line]))
                return current

            previous = token

        raise Exception("unexpected EOF on line {}: '{}'".format(
            token.line + 1, self.lines[token.line]))

    def _parse_field(self, field, tokens):
        token = next(tokens)

        if token.token_type == 'LBRACKET':
            for token in tokens:
                if token.token_type == 'DEFAULT':
                    field.default = token.value
                elif token.token_type == 'PACKED':
                    field.packed = token.value
                elif token.token_type == 'DEPRECATED':
                    field.deprecated = token.value
                elif token.token_type == 'CUSTOM':
                    # Ignore custom modifiers for now
                    continue
                elif token.token_type == 'COMMA':
                    continue
                else:
                    assert token.token_type == 'RBRACKET', (
                        "unexpected {} token on line {}: '{}'".format(
                            token.token_type, token.line + 1,
                            self.lines[token.line]))
                    return

        else:
            assert token.token_type == 'SEMICOLON', (
                "expected ; or modifier on line {}, got {}: '{}'".format(
                    token.line + 1, self.lines[token.line],
                    token.token_type))

    def _parse_enum(self, current, tokens):
        token = next(tokens)
        assert token.token_type == 'LBRACE', (
            "missing opening paren on line {}: '{}'".format(
                token.line + 1, self.lines[token.line]))

        for token in tokens:
            if token.token_type == 'ENUM_FIELD':
                token.full_name = "%s_%s" % (current.full_name, token.name)
                self._parse_enum_field(token, tokens)
                current.fields.append(token)

            else:
                assert token.token_type == 'RBRACE', (
                    "unexpected %s token on line {}: '{}'".format(
                        token.token_type, token.line + 1,
                        self.lines[token.line]))
                return current

        raise Exception("unexpected EOF on line {}: '{}'".format(
            token.line + 1, self.lines[token.line]))

    def _parse_enum_field(self, field, tokens):
        token = next(tokens)

        if token.token_type == 'LBRACKET':
            for token in tokens:
                if token.token_type == 'CUSTOM':
                    # Ignore custom modifiers for now
                    continue
                elif token.token_type == 'COMMA':
                    continue
                else:
                    assert token.token_type == 'RBRACKET', (
                        "unexpected {} token on line {}: '{}'".format(
                            token.token_type, token.line + 1,
                            self.lines[token.line]))
                    return

        else:
            assert token.token_type == 'SEMICOLON', (
                "expected ; or modifier on line {}, got {}: '{}'".format(
                    token.line + 1, token.token_type,
                    self.lines[token.line]))

    def _parse_extend(self, current, tokens):
        token = next(tokens)
        assert token.token_type == 'LBRACE', (
            "missing opening paren on line {}: '{}'".format(
                token.line + 1, self.lines[token.line]))

        # For now, just find the closing brace and return
        for token in tokens:
            if token.token_type == 'RBRACE':
                return current

    def add_cython_info(self, message):
        for field in message.fields:
            field.list_type = self.list_type_map.get(field.type, 'TypedList')
            field.fixed_width = (field.type in {
                'float', 'double', 'fixed32', 'sfixed32', 'fixed64', 'sfixed64'
            })
            field.var_width = (field.type in {
                'bool', 'enum', 'int32', 'sint32', 'uint32', 'int64', 'sint64',
                'uint64'
            })

            if field.type in self.scalars:
                field.c_type = self.c_type_map[field.type]

            if field.var_width:
                field.getter = self.getter_map[field.type]
                field.setter = self.setter_map[field.type]

        for submessage in message.messages.values():
            self.add_cython_info(submessage)

    class Token(object):
        token_type = None
        line = -1
        type = None

    class Option(Token):
        def __init__(self, line, option):
            self.token_type = 'OPTION'
            self.line = line
            self.option = option

    class Syntax(Token):
        def __init__(self, line, value):
            self.token_type = 'SYNTAX'
            self.line = line
            self.value = value

    class Import(Token):
        def __init__(self, line, value):
            self.token_type = 'IMPORT'
            self.line = line
            self.value = value

    class Message(Token):
        def __init__(self, line, name):
            self.token_type = 'MESSAGE'
            self.line = line
            self.name = name
            # full_name may later be overridden with parent hierarchy
            self.full_name = name
            self.messages = {}
            self.enums = {}
            self.fields = []

    class Modifier(Token):
        def __init__(self, line, value):
            self.token_type = 'MODIFIER'
            self.line = line
            self.value = value

    class Field(Token):
        def __init__(self, line, ftype, name, index):
            self.token_type = 'FIELD'
            self.line = line
            self.modifier = None
            self.type = ftype
            self.name = name
            self.index = int(index)
            self.default = None
            self.packed = False
            self.deprecated = False

        def get_key(self):
            if self.modifier == 'repeated' and self.packed:
                return (self.index << 3) | 2
            elif self.type in {'message', 'string', 'bytes'}:
                return (self.index << 3) | 2
            elif self.type in {'fixed64', 'sfixed64', 'double'}:
                return (self.index << 3) | 1
            elif self.type in {'fixed32', 'sfixed32', 'float'}:
                return (self.index << 3) | 5
            else:
                return (self.index << 3) | 0

    class LBracket(Token):
        def __init__(self, line):
            self.token_type = 'LBRACKET'
            self.line = line

    class RBracket(Token):
        def __init__(self, line):
            self.token_type = 'RBRACKET'
            self.line = line

    class Default(Token):
        def __init__(self, line, value):
            self.token_type = 'DEFAULT'
            self.line = line
            self.value = process_default(value)

    class Packed(Token):
        def __init__(self, line, value):
            self.token_type = 'PACKED'
            self.line = line
            self.value = value == 'true'

    class Deprecated(Token):
        def __init__(self, line, value):
            self.token_type = 'DEPRECATED'
            self.line = line
            self.value = value == 'true'

    class Custom(Token):
        def __init__(self, line, value):
            self.token_type = 'CUSTOM'
            self.line = line
            self.value = value

    class Comma(Token):
        def __init__(self, line):
            self.token_type = 'COMMA'
            self.line = line

    class Semicolon(Token):
        def __init__(self, line):
            self.token_type = 'SEMICOLON'
            self.line = line

    class Enum(Token):
        def __init__(self, line, name):
            self.token_type = 'ENUM'
            self.line = line
            self.name = name
            self.fields = []
            # full_name may later be overridden with parent hierarchy
            self.full_name = name

    class EnumField(Token):
        def __init__(self, line, name, value=None):
            self.token_type = 'ENUM_FIELD'
            self.line = line
            self.name = name
            if value is not None:
                self.value = int(value, 0)
            else:
                self.value = None
            # full_name may later be overridden with parent hierarchy
            self.full_name = name

    class LBrace(Token):
        def __init__(self, line):
            self.token_type = 'LBRACE'
            self.line = line

    class RBrace(Token):
        def __init__(self, line):
            self.token_type = 'RBRACE'
            self.line = line

    class Extension(Token):
        def __init__(self, line, low, hi):
            self.token_type = 'EXTENSION'
            self.line = line
            self.low = low
            self.hi = hi

    class Oneof(Token):
        def __init__(self, line, name):
            self.token_type = 'ONEOF'
            self.line = line
            self.name = name

    class Extend(Token):
        def __init__(self, line, name):
            self.token_type = 'EXTEND'
            self.line = line
            self.name = name


class Proto3Parser(Parser):
    syntax = 3


def process_default(default):
    if default == 'true':
        return True
    elif default == 'false':
        return False
    else:
        return default



"""Setuptools integration."""
from pyrobuf.compile import Compiler

try:
    basestring
except NameError:
    # Python 3.x
    basestring = str


def add_pyrobuf_module(dist, pyrobuf_module):
    dir_name = "pyrobuf/_" + pyrobuf_module
    compiler = Compiler(pyrobuf_module, out=dir_name,
                        package="{}_{}".format(dist.get_name(), pyrobuf_module))
    compiler.extend(dist)


def pyrobuf_modules(dist, attr, value):
    assert attr == 'pyrobuf_modules'
    if isinstance(value, basestring):
        value = [value]

    for pyrobuf_module in value:
        add_pyrobuf_module(dist, pyrobuf_module)



from libc.stdint cimport *

cdef safe_fill(char *dest, const unsigned char *source, size_t max_len)

cdef safe_dup(const unsigned char *source, size_t max_len)

cdef int32_t get_varint32(const unsigned char *memory, int *offset)

cdef int64_t get_varint64(const unsigned char *memory, int *offset)

cdef int32_t get_signed_varint32(const unsigned char *memory, int *offset)

cdef int64_t get_signed_varint64(const unsigned char *memory, int *offset)

cdef int set_varint32(int32_t varint, bytearray buf)

cdef int set_varint64(int64_t varint, bytearray buf)

cdef int set_signed_varint32(int32_t varint, bytearray buf)

cdef int set_signed_varint64(int64_t varint, bytearray buf)

cdef bint skip_generic(const unsigned char *memory, int *offset, int size, int wire_type)
